// header
#include <Obj.H>
#include <Misc.H>
#include <Game.H>

/* ##############################################################
   Part
   ############################################################## */

List <Part *> Part::partList;
int Part::idCount = 0;
int Part::numofPart = 0;

Part::Part(Part *p, const XsImage *i, const DVector &b, const DVector &l, const DVector &j, const double de,const double di) : XsSprite(::view,i,b,de,1), parent(p), id(idCount++), pixmapRound(NULL), direct(di), reldirect(di), relbase(b), rellocate(l), reljoint(j) {
    if (parent) {
	parent->childList.add(this);
	flushPart(parent->locate, parent->direct);
    } else {
	partList.add(this);
    }
    numofPart ++;
}

Part::Part(Part *p, const XsPixmapRound *i, const DVector &b, const DVector &l, const DVector &j, const double de,const double di) : XsSprite(::view,i,b,de,1), parent(p), id(idCount++), pixmapRound(i), direct(di), reldirect(di), relbase(b), rellocate(l), reljoint(j) {
    if (parent) {
	parent->childList.add(this);
	flushPart(parent->locate, parent->direct);	
    } else {
	partList.add(this);
    }
    numofPart ++;
}

Part::~Part(void) {
    if (parent) {
	parent->childList.remove(this);
    } else {
	partList.remove(this);
    }
    WriteIter <Part *> wi(childList);
    while (wi.existNext()) {
	delete wi.elementNext();
    }
    numofPart --;        
}

void Part::flushPart(const DVector &l, const double d) {
    direct = d + reldirect;
    base = l + relbase % d;
    locate = base + rellocate % direct;
    joint = locate + reljoint % direct;
    if (pixmapRound) {
	image = &pixmapRound->image(direct);
    }
    WriteIter <Part *> wi(childList);
    while (wi.existNext()) {
	wi.elementNext()->flushPart(joint,direct);
	wi.next();
    }
}

void Part::flushAllPart(void) {
    WriteIter <Part *> wp(partList);
    while (wp.existNext()) {
	Part *p = wp.elementNext();
	if (p->pixmapRound) {
	    p->image = &p->pixmapRound->image(p->direct);
	}
	WriteIter <Part *> wc(p->childList);
	while (wc.existNext()) {
	    Part *c = wc.elementNext();
	    c->flushPart(p->locate,p->direct);
	    wc.next();
	}
	wp.next();
    }
}

void Part::setDirect(const DVector &l, const double s, const double r) {
    while (!adjustDirect(l,s,r)) {}
}

int Part::adjustDirect(const DVector &l, const double s, const double r) {
    const double d = direct + r;
    const Vector v = (l-locate) % (-d);
    if (v.x < 0) {
	if (0 <= v.y) {
	    direct += s;
	} else {
	    direct -= s;
	}
	return 0;
    } else {
	if (0 < v.y) {
	    const Vector vd = (l-locate) % -(d + s);
	    if (0 < vd.y) {
		direct += s;
		return 0;
	    } else {
		return 1;		
	    }
	} 
	if (v.y < 0) {
	    const Vector vd = (l-locate) % -(d - s);
	    if (vd.y < 0) {
		direct -= s;
		return 0;
	    } else {
		return 1;
	    }
	}
	return 1;
    }
}
    
/* ##################################################################
   Object
   ################################################################## */

List <Obj *> Obj::objList;
const Obj::Attr Obj::attr(1, DVector(1, 1), 1, 1, 0);

Obj::Obj(Part *p, const XsImage *i, const DVector &b, const DVector &l, const DVector &j, const double de, const double di) : Part(p,i,b,l,j,de,di), isAlive(1), damage(0) {
    objList.add(this);
}

Obj::Obj(Part *p, const XsPixmapRound *i, const DVector &b, const DVector &l, const DVector &j, const double de, const double di) : Part(p,i,b,l,j,de,di), isAlive(1), damage(0) {
    objList.add(this);
}

Obj::~Obj(void) {
    objList.remove(this);
}

const int Obj::inView(void) const {
    return locate.in(view.size());
}

const int Obj::isOverlap(const Obj *o) const {
    
    Vector st = locate		+ getAttr().range	* image->top();
    Vector sb = locate		+ getAttr().range	* image->bottom();
    Vector dt = o->locate	+ o->getAttr().range	* o->image->top();
    Vector db = o->locate	+ o->getAttr().range	* o->image->bottom();
    if (db.x < st.x || sb.x < dt.x) return 0;
    if (db.y < st.y || sb.y < dt.y) return 0;
    return 1;
}

void Obj::onFlush(void) {}

void Obj::onDestroy(void) {}

void Obj::flushAllObj(void) {
    
    // ֥Ȥν
    WriteIter <Obj *> ri(objList);
    ri.head();    
    while (ri.existNext()) {
	Obj *o = ri.elementNext();
	// ġν
	o->onFlush();
	// ¸Ƚ
	if (!o->isAlive) {
	    delete o;
	    // ˲ȼưŪ˥ꥹȤ롥
	    continue;
	}	
	ri.next();
    }
}

const Obj::Attr &Obj::getAttr(void) const { return attr; }

Obj::Attr::Attr(unsigned int m, const DVector &r, unsigned int o, unsigned int d, unsigned int p) : isMatter(m), range(r), offense(o), defense(d), point(p) {}

/* ##################################################################
   Ship
   ################################################################## */

Ship *Ship::_ship = NULL;

Ship::Ship(Part *p, const XsImage *i, const DVector &b, const DVector &l, const DVector &j, const double de, const double di, const int m) : Obj(p,i,b,l,j,de,di), _subWeapon(m) {
    if (_ship) {
	cerr << "Error !!! : Ship object is already existing." << endl;
	exit(1);
    }
    _ship = this;
}

Ship::Ship(Part *p, const XsPixmapRound *i,const DVector &b, const DVector &l, const DVector &j, const double de, const double di, const int m) : Obj(p,i,b,l,j,de,di), _subWeapon(m) {
    if (_ship) {
	cerr << "Error !!! : Ship object is already existing." << endl;
	exit(1);
    }
    _ship = this;
}

Ship::~Ship(void) {
    _ship = NULL;
}

Ship *Ship::ship(void) {
    return _ship;
}

const int Ship::subWeapon(void) const {
    return _subWeapon;
}

int Ship::useMainWeapon(void) {
    return 0;
}

int Ship::addSubWeapon(void) {
    return 0;
}

int Ship::removeSubWeapon(void) {
    return 0;
}

/* ##################################################################
   ShipAttack
   ################################################################## */

List <ShipAttack *> ShipAttack::shipAttackList;

ShipAttack::ShipAttack(Part *p, const XsImage *i, const DVector &b, const DVector &l, const DVector &j, const double de, const double di) : Obj(p,i,b,l,j,de,di) {
    shipAttackList.add(this);
}

ShipAttack::ShipAttack(Part *p, const XsPixmapRound *i, const DVector &b, const DVector &l, const DVector &j, const double de, const double di) : Obj(p,i,b,l,j,de,di) {
    shipAttackList.add(this);
}

ShipAttack::~ShipAttack(void) {
    shipAttackList.remove(this);        
}

void ShipAttack::checkAttack(void) {
    
    if (shipAttackList.num() && Enemy::enemyList.num()) {
	WriteIter <Enemy *> ei(Enemy::enemyList);
	while (ei.existNext()) {
	    Obj *e = ei.elementNext();
	    if (e->inView()) {
		WriteIter <ShipAttack *> ai(ShipAttack::shipAttackList);
		while (ai.existNext()) {
		    Obj *a = ai.elementNext();
		    if (a->inView() && e->isOverlap(a)) {
			// Ũ˥᡼ɲ
			e->damage += a->getAttr().offense;
			int d = (e->getAttr().defense <= e->damage);
			int c = d ? 0 : 5;
			// 
			if (a->getAttr().offense) {
			    DVector v = DVector(localRand(9)-4,
						localRand(9)-4);
			    new Light(0, e->locate + v, c);
			}
			// ˲Ƚ
			if (a->getAttr().isMatter) {
			    // ˲
			    a->onDestroy();
			    delete a;
			    a = NULL;
			}
			// Ũ˲Ƚ
			if (d) {
			    // û
			    Game::game()->score += e->getAttr().point;
			    // ˲
			    e->onDestroy();
			    delete e;
			    e = NULL;
			    break;
			}
		    }
		    if (a) ai.next();
		}
	    }
	    if (e) ei.next();
	}
    }
    
}
		     
/* ##################################################################
   Enemy
   ################################################################## */

List <Enemy *> Enemy::enemyList;

Enemy::Enemy(Part *p, const XsImage *i, const DVector &b, const DVector &l, const DVector &j, const double de, const double di) : Obj(p,i,b,l,j,de,di) {
    enemyList.add(this);
}

Enemy::Enemy(Part *p, const XsPixmapRound *i, const DVector &b, const DVector &l, const DVector &j, const double de, const double di) : Obj(p,i,b,l,j,de,di) {
    enemyList.add(this);
}

Enemy::~Enemy(void) {
    enemyList.remove(this);        
}

void Enemy::checkAttack(void) {
    
    if (enemyList.num() && Ship::ship()) {
	Ship *s = Ship::ship();
	if (s->inView()) {
	    WriteIter <Enemy *> ai(Enemy::enemyList);
	    ai.head();
	    while (ai.existNext()) {
		Obj *a = ai.elementNext();
		if (a->inView() && s->isOverlap(a)) {
		    // ˥᡼ɲ
		    s->damage += a->getAttr().offense;
		    a->damage += s->getAttr().offense;		    
		    int sd = (s->getAttr().defense <= s->damage);
		    int sc = sd ? 0 : 5;
		    int ad = (a->getAttr().defense <= a->damage);
		    // 
		    if (a->getAttr().offense) {
			for (int i = 0; i < 5; i ++) {
			    DVector v = DVector(localRand(9)-4,
						localRand(9)-4);
			    new Light(i, s->locate + v, sc);
			}
		    }
		    // ˲Ƚ
		    if (ad && a->getAttr().isMatter) {
			// ˲
			a->onDestroy();
			delete a;
			a = NULL;
		    }
		    // ˲Ƚ
		    if (sd) {
			// û
			Game::game()->score += s->getAttr().point;
			// ˲
			s->onDestroy();
			delete s;
			s = NULL;
			break;
		    }
		}
		if (a) ai.next();
	    }
	}
    }
}
		     
/* ##################################################################
   EnemyAttack
   ################################################################## */

List <EnemyAttack *> EnemyAttack::enemyAttackList;

EnemyAttack::EnemyAttack(Part *p, const XsImage *i, const DVector &b, const DVector &l, const DVector &j, const double de, const double di) : Obj(p,i,b,l,j,de,di) {
    enemyAttackList.add(this);
}

EnemyAttack::EnemyAttack(Part *p, const XsPixmapRound *i, const DVector &b, const DVector &l, const DVector &j, const double de, const double di) : Obj(p,i,b,l,j,de,di) {
    enemyAttackList.add(this);
}

EnemyAttack::~EnemyAttack(void) {
    enemyAttackList.remove(this);        
}

void EnemyAttack::checkAttack(void) {
    
    if (enemyAttackList.num() && Ship::ship()) {
	Ship *s = Ship::ship();
	if (s->inView()) {
	    WriteIter <EnemyAttack *> ai(EnemyAttack::enemyAttackList);
	    ai.head();
	    while (ai.existNext()) {
		Obj *a = ai.elementNext();
		if (a->inView() && s->isOverlap(a)) {
		    // ˥᡼ɲ
		    s->damage += a->getAttr().offense;
		    int d = (s->getAttr().defense <= s->damage);
		    int c = d ? 0 : 5;
		    // 
		    if (a->getAttr().offense) {
			for (int i = 0; i < 5; i ++) {
			    DVector v = DVector(localRand(9)-4,
						localRand(9)-4);
			    new Light(i, s->locate + v, c);
			}
		    }
		    // ˲Ƚ
		    if (a->getAttr().isMatter) {
			// ˲
			a->onDestroy();
			delete a;
			a = NULL;
		    }
		    // ˲Ƚ
		    if (d) {
			// û
			Game::game()->score += s->getAttr().point;
			// ˲
			s->onDestroy();
			delete s;
			s = NULL;
			break;
		    }
		}
		if (a) ai.next();
	    }
	}
    }
}
		     
/* ##################################################################
   Misc
   ################################################################## */

List <Misc *> Misc::miscList;

Misc::Misc(Part *p, const XsImage *i, const DVector &b, const DVector &l, const DVector &j, const double de, const double di) : Obj(p,i,b,l,j,de,di) {
    miscList.add(this);
}

Misc::Misc(Part *p, const XsPixmapRound *i, const DVector &b, const DVector &l, const DVector &j, const double de, const double di) : Obj(p,i,b,l,j,de,di) {
    miscList.add(this);
}

Misc::~Misc(void) {
    miscList.remove(this);        
}

/* ##################################################################
   HaveTarget
   ################################################################## */

HaveTarget::HaveTarget(Obj *o) {
    setTarget(o);
}

int HaveTarget::setTarget(Obj *o) {
    if (Obj::objList.exist(o)) {
	// ¸ߤ
	objTarget = o;
	idTarget = o->id;
	return 1;
    } else {
	// ¸ߤʤä
	objTarget = NULL;
	idTarget = 0;
	return 0;
    }
}

int HaveTarget::updateTarget(void) {
    if (!objTarget) {
	// ¸ߤʤ
	return 0;
    } else if (!Obj::objList.exist(objTarget) || objTarget->id != idTarget) {
	// ¸ߤʤ⤷ֹ㤤ξ
	objTarget = NULL;
	return 0;
    } else {
	return 1;
    }
}
