#ifndef Obj_H
#define Obj_H

// header
#include <limits.h>

#include <DVector.H>
#include <List.H>
#include <XsSprite.H>
#include <XsPixmapRound.H>

#include <common.H>

/* ##################################################################
   Part
   ################################################################## */

class Part : public XsSprite {

  public:
    
    Part(Part *,
	 const XsImage *,
	 const DVector & = dzero,
	 const DVector & = dzero, 
	 const DVector & = dzero, 
	 const double = 0, const double = 0);
    
    Part(Part *,
	 const XsPixmapRound *,
	 const DVector & = dzero,	 
	 const DVector & = dzero,
	 const DVector & = dzero,
	 const double = 0, const double = 0);
    virtual ~Part(void);

    Part * const parent;		// °
    List <Part *> childList;		// ҤΥꥹ
    
    int id;				// ֹ
    
    const XsPixmapRound *pixmapRound;	// žԥåޥå
    
    double direct;			// 
    double reldirect;			// 

    DVector base;			// е
    DVector joint;			// Ϣ
    
    DVector relbase;			// е
    DVector rellocate;			// а
    DVector reljoint;			// Ϣ

    void setDirect(const DVector &,
		   const double = 10,
		   const double = 0);	// 
    int adjustDirect(const DVector &,	// Ĵ
		     const double = 10,
		     const double = 0);
    
    void flushPart(const DVector &,	// ʬη׻
		   const double);

    static void flushAllPart(void);	// ʤη׻

    static List <Part *> partList;	// ƥѡĤΥꥹ
    static int idCount;			// ֹΥ

    static int numofPart;		// ºݤΥѡĤο
};

/* ##################################################################
   Object
   ################################################################## */

class Obj : public Part {

public:

    Obj(Part *,
	const XsImage *,
	const DVector & = dzero,
	const DVector & = dzero,
	const DVector & = dzero,
	const double = 0, const double = 0);
    Obj(Part *,
	const XsPixmapRound *,
	const DVector & = dzero,
	const DVector & = dzero,
	const DVector & = dzero,
	const double = 0, const double = 0);
    virtual ~Obj(void);

    int isAlive;			// ̵ͭ

    /* ######################################################
       ֥Ȥ°
       ###################################################### */
    
    struct Attr {
	
    public:

	// ֥Ȥü
	Attr(unsigned int isMatter,
	     const DVector &range,
	     unsigned int offense,
	     unsigned int defense,
	     unsigned int point);
	
	const unsigned int isMatter;
	const DVector range;
	const unsigned int offense;
	const unsigned int defense;
	const unsigned int point;
    };
    
    unsigned int damage;				// ᡼
    static const Attr attr;				// °
    virtual const Obj::Attr &getAttr(void) const;	// °μ
    
    const int inView(void) const;
    virtual const int isOverlap(const Obj *) const;

    virtual void onFlush(void);

    virtual void onDestroy(void);
    
    static void flushAllObj(void);
    
    static List <Obj *> objList;
};

/* ##################################################################
   Enemy
   ################################################################## */

class Enemy : public Obj {
    
public:
    Enemy(Part *,
	  const XsImage *,
	  const DVector & = dzero,
	  const DVector & = dzero,
	  const DVector & = dzero,
	  const double = 0, const double = 0);
    Enemy(Part *,
	  const XsPixmapRound *,
	  const DVector & = dzero,
	  const DVector & = dzero,
	  const DVector & = dzero,
	  const double = 0, const double = 0);
    virtual ~Enemy(void);

    static void checkAttack(void);
    static List <Enemy *> enemyList;
};

/* ##################################################################
   EnemyAttack
   ################################################################## */

class EnemyAttack : public Obj {
    
public:

    EnemyAttack(Part *,
		const XsImage *,
		const DVector & = dzero,
		const DVector & = dzero,
		const DVector & = dzero,
		const double = 0, const double = 0);
    EnemyAttack(Part *,
		const XsPixmapRound *,
		const DVector & = dzero,
		const DVector & = dzero,
		const DVector & = dzero,
		const double = 0, const double = 0);
    virtual ~EnemyAttack(void);
	  
    static void checkAttack(void);
    static List <EnemyAttack *> enemyAttackList;
};

/* ##################################################################
   Ship
   ################################################################## */

class Ship : public Obj {
    
public:

    Ship(Part *,
	 const XsImage *,
	 const DVector & = dzero,
	 const DVector & = dzero,
	 const DVector & = dzero,
	 const double = 0,
	 const double = 0,
	 const int = 0);
    Ship(Part *,
	 const XsPixmapRound *,
	 const DVector & = dzero,
	 const DVector & = dzero,
	 const DVector & = dzero,
	 const double = 0,
	 const double = 0,
	 const int = 0);
    virtual ~Ship(void);

    static Ship *ship(void);

    virtual int useMainWeapon(void);
    virtual int addSubWeapon(void);
    virtual int removeSubWeapon(void);

    const int subWeapon(void) const;
    
  protected:
    int _subWeapon;
    
private:
    static Ship *_ship;
};

/* ##################################################################
   ShipAttack
   ################################################################## */

class ShipAttack : public Obj {
    
public:

    ShipAttack(Part *,
	       const XsImage *,
	       const DVector & = dzero,
	       const DVector & = dzero,
	       const DVector & = dzero,
	       const double = 0, const double = 0);
    ShipAttack(Part *,
	       const XsPixmapRound *,
	       const DVector & = dzero,
	       const DVector & = dzero,
	       const DVector & = dzero,
	       const double = 0, const double = 0);
    virtual ~ShipAttack(void);
	  
    static void checkAttack(void);
    static List <ShipAttack *> shipAttackList;
};

/* ##################################################################
   Misc
   ################################################################## */

class Misc : public Obj {
    
public:

    Misc(Part *,
	 const XsImage *,
	 const DVector & = dzero,
	 const DVector & = dzero,
	 const DVector & = dzero,
	 const double = 0, const double = 0);
    Misc(Part *,
	 const XsPixmapRound *,
	 const DVector & = dzero,
	 const DVector & = dzero,
	 const DVector & = dzero,
	 const double = 0, const double = 0);
    virtual ~Misc(void);
	  
    static List <Misc *> miscList;
};

/* ###############################################################
   HaveTarget
   ############################################################### */

class HaveTarget {
    
  public:
    HaveTarget(Obj *);

  protected:
    int setTarget(Obj *);
    int updateTarget(void);
    
    Obj *objTarget;
    int idTarget;
};

#endif


