// إå
#include <XsBitmap.H>

//
// XsBitmap.C
//

// Ū
const int XsBitmap::xMax = 64;
const int XsBitmap::yMax = 64;

// 󥹥ȥ饯
XsBitmap::XsBitmap(const String &f,const XsColor &c)
: _bitmap(0),pixel(c.color.pixel),oldbitmap(0),oldpixel(0) {
    create(f);
}
XsBitmap::XsBitmap(const XsBitmap &p)
: _bitmap(0),pixel(p.pixel),oldbitmap(0),oldpixel(0) {
    create(p);
}

// ǥȥ饯
XsBitmap::~XsBitmap(void) {
    destroy();
}

// ˲
void XsBitmap::destroy(void) {

    if (_bitmap) { XFreePixmap(display,_bitmap); _bitmap = 0; }
    XFlush(display);
    
    _size = zero;
    _hots = zero;
}

void XsBitmap::color(const XsColor &c) {

    pixel = c.color.pixel;
}

// 
void XsBitmap::create(const Vector &vec) {

    destroy();

    {
	_bitmap = XCreatePixmap(display, root, vec.x, vec.y, 1);
	XFillRectangle(display, _bitmap, gcWhite, 0, 0, vec.x, vec.y);
    }
    
    _size = vec;
    _hots = zero;
}

void XsBitmap::create(const String &fileName) {

    destroy();

    {
	int err = XReadBitmapFile(display,root,
				  fileName,
				  (unsigned int *)&_size.x,
				  (unsigned int *)&_size.y,
				  &_bitmap,
				  &_hots.x,&_hots.y);

	if (err != BitmapSuccess) {
	    cerr << "ERROR !!! : cannot open \"" << fileName << "\"." << endl;
	    exit(1);
	}

	if (xMax < _size.x || yMax < _size.y) {
	    cerr << "WARNING !!! : too big bitmap" << endl;
	}

	_size = Vector((_size.x < xMax) ? _size.x : xMax,
		       (_size.y < yMax) ? _size.y : yMax);
    }
}

void XsBitmap::create(const XsBitmap &p) {

    destroy();

    {
	_size = p._size;
	_hots = p._hots;
	
	_bitmap = XCreatePixmap(display, root,
				_size.x, _size.y, 1);
	XCopyArea(display,p._bitmap,_bitmap,gc1Copy,0,0,_size.x,_size.y,0,0);
	XFlush(display);
    }
}

void XsBitmap::draw(const XsPixmap &p,const Vector &v) const {

    int xx = v.x - _hots.x;
    int yy = v.y - _hots.y;
    
    XSetTSOrigin(display, gcStippled, xx, yy);
    XSetForeground(display,gcStippled,pixel);
    XSetStipple(display,gcStippled,_bitmap);
    XFillRectangle(display, p.pixmap() ,gcStippled,
		   xx, yy ,_size.x, _size.y);
}

Pixmap XsBitmap::bitmap(void) const {
    return _bitmap;
}

// 
int XsBitmap::isChanged(void) const {

    return (oldbitmap != _bitmap ||
	    oldpixel != pixel ||
	    XsImage::isChanged());
}

void XsBitmap::flushImage(void) {

    oldbitmap = _bitmap;
    oldpixel = pixel;
    XsImage::flushImage();
}

