// إå
#include <XsPixmap.H>

/* ######################################################
   class XsPixmap
   ###################################################### */

const int XsPixmap::xMax = 1024;
const int XsPixmap::yMax = 1024;

XsPixmap::XsPixmap(void)
: _pixmap(0),_mask(0),oldpixmap(0) {}

XsPixmap::XsPixmap(const Vector &vec,const Vector &h)
: _pixmap(0),_mask(0),oldpixmap(0) {
    create(vec,h);
}
XsPixmap::XsPixmap(const String &f)
: _pixmap(0),_mask(0),oldpixmap(0) {
    create(f);
}
XsPixmap::XsPixmap(const XsPixmap &p)
: _pixmap(0),_mask(0),oldpixmap(0) {
    create(p);
}

XsPixmap::~XsPixmap(void) {
    destroy();
}

void XsPixmap::destroy(void) {

    if (_pixmap)	{ XFreePixmap(display,_pixmap);	_pixmap = 0; }
    if (_mask)		{ XFreePixmap(display,_mask);	_mask = 0; }
    
    XFlush(display);
    
    _size = zero;
    _hots = zero;
}

void XsPixmap::create(const Vector &vec,const Vector &h) {

    destroy();

    {
	_pixmap = XCreatePixmap(display, root, vec.x, vec.y, depth);
	_mask = XCreatePixmap(display, root, vec.x, vec.y, 1);
	
	XFillRectangle(display, _pixmap, gcBlack, 0, 0, vec.x, vec.y);
	XFillRectangle(display, _mask, gc1Black, 0, 0, vec.x, vec.y);
	
	XFlush(display);
    }
    
    _size = vec;
    _hots = h;
}

void XsPixmap::create(const String &f) {

    destroy();

    {
	XpmAttributes attr;
	attr.valuemask = 0;
	attr.x_hotspot = 0;
	attr.y_hotspot = 0;

	/* char fn[] ѤƤΤ
	   ΤXpmRead...()3 const char * Ǥʤ顥 */
	char fn[f.length()+1];
	strcpy(fn,f);
	
	int err = XpmReadFileToPixmap(display, root, fn,
				      &_pixmap,&_mask,&attr);
	switch (err) { 
	  case XpmOpenFailed:
	    cerr << "Error !!! : cannot open \"" << f << "\"." << endl;
	    exit(1);
	  case XpmFileInvalid:
	    cerr << "Error !!! : invalid xpm file \"" << f << "\"" << endl;
	    exit(1);
	  case XpmNoMemory:
	    cerr << "Error !!! : not enough memory" << endl;
	    exit(1);
	  case XpmColorFailed:
	    cerr << "Error !!! : can\'t get any colors" << endl;
	    exit(1);
	  default:
	    break;
	}

	if (xMax < attr.width || yMax < attr.height) {
	    cerr << "WARNING !!! : too big pixmap" << endl;
	}

	_size = Vector((attr.width  < xMax) ? attr.width  : xMax,
		       (attr.height < yMax) ? attr.height : yMax);
	_hots = Vector(attr.x_hotspot,
		       attr.y_hotspot);

	XFlush(display);
    }
}

void XsPixmap::create(const XsPixmap &p) {

    destroy();

    {
	_size = p._size;
	_hots = p._hots;
	
	_pixmap = XCreatePixmap(display, root,_size.x, _size.y, depth);
	XCopyArea(display,p._pixmap,_pixmap,gcCopy,0,0,_size.x,_size.y,0,0);
	
	if (p._mask) {
	    _mask = XCreatePixmap(display, root,_size.x, _size.y, 1);
	    XCopyArea(display,p._mask,_mask,gc1Copy,0,0,_size.x,_size.y,0,0);
	}
	
	XFlush(display);
    }
}

void XsPixmap::addMask(void) {
    if (!_mask) {
	_mask = XCreatePixmap(display, root, _size.x, _size.y, 1);
	XFillRectangle(display, _mask, gc1White, 0, 0, _size.x, _size.y);
	XFlush(display);
    }
}

void XsPixmap::draw(const XsPixmap &p,const Vector &v) const {

    int xx = v.x - _hots.x;
    int yy = v.y - _hots.y;

    if (_mask) {
	XCopyPlane(display,
		   _mask, p._pixmap, gcAnd,
		   0, 0, _size.x, _size.y, xx, yy, 1);
	XCopyArea(display,
		  _pixmap, p._pixmap, gcOr,
		  0, 0, _size.x, _size.y, xx, yy);
    } else {
	XCopyArea(display,
		  _pixmap,p._pixmap, gcCopy,
		  0, 0, _size.x, _size.y, xx, yy);
    }
}

Pixmap XsPixmap::pixmap(void) const {
    return _pixmap;
}

Pixmap XsPixmap::mask(void) const {
    return _mask;
}

int XsPixmap::isChanged(void) const {
    return (oldpixmap != _pixmap ||
	    XsSolid::isChanged());
}

void XsPixmap::flushImage(void) {
    oldpixmap = _pixmap;
    XsSolid::flushImage();
}
