// header
#include <iomanip.h>

#include <string.h>
#include <XsView.H>

/* #####################################################################
   struct SpriteContena
   ##################################################################### */

struct SpriteContena {
    XsSprite *sprite;
    Vector top;
    Vector bottom;
};

/* #####################################################################
   class DepthSortedContena
   ##################################################################### */

class DepthSortedContena : public List<SpriteContena *> {

// 
public:
    virtual void add(SpriteContena *c) {
        WriteIter<SpriteContena *> i(*(List<SpriteContena *>*)this);
        i.head();
        while (i.existNext()) {
	    const SpriteContena *e = i.elementNext();
	    double cd = c->sprite->depth;
	    double ed = e->sprite->depth;
            if (cd < ed || cd == ed && c < e) {
                break;
            }
            i.next();
        }
        i.insertNext(c);
    }
};

/* #####################################################################
   class DepthSortedSprite
   ##################################################################### */

class DepthSortedSprite : public List<XsSprite *> {

// 
public:
    virtual void add(XsSprite *c) {
        WriteIter<XsSprite *> i(*(List<XsSprite *>*)this);
        i.head();
        while (i.existNext()) {
	    const XsSprite *e = i.elementNext();
	    double cd = c->depth;
	    double ed = e->depth;
            if (cd < ed || cd == ed && c < e) {
                break;
            }
            i.next();
        }
        i.insertNext(c);
    }
};

/* ###################################################################
   class XsView
   ################################################################### */

// Ū
Set <XsView *> XsView::setView;

XsView::XsView(const Window &ww, const Vector &vec, int r)
: _redraw(r) ,window(ww) ,tileSize(vec),  isFillUnchangedTiles(0), _numChangedTiles(0), _numChangedAreas(0) {

    // ؤɲ
    setView.add(this);
    
    // ɥ礭
    XWindowAttributes attr;
    XGetWindowAttributes(display,window,&attr);

    // ɥ礭ǥԥåޥåפΤ
    create(Vector(attr.width, attr.height));

    // ޥå
    const XsImage **tim;
    int *tit;
    
    _mapSize	= _size / tileSize;
    mapArray	= _mapSize.x * _mapSize.y;
    
    _map	= new (const XsImage **)[_mapSize.y];
    tim		= new (const XsImage *)[mapArray];
    for (int i = 0;i < _mapSize.y;i ++) { _map[i] = tim + i * _mapSize.x; }
    memset(_map[0],int(NULL),sizeof(const XsImage *) * mapArray);
		       
    _oldmap	= new (const XsImage **)[_mapSize.y];
    tim		= new (const XsImage *)[mapArray];
    for (int i = 0;i < _mapSize.y;i ++) { _oldmap[i] = tim + i * _mapSize.x; }
    memset(_oldmap[0],int(NULL),sizeof(const XsImage *) * mapArray);
		       
    _state	= new (int *)[_mapSize.y];
    tit		= new int[mapArray];
    for (int i = 0;i < _mapSize.y;i ++) { _state[i] = tit + i * _mapSize.x; }
    memset(_state[0],0,sizeof(int) * mapArray);
}

XsView::~XsView(void) {

    // 礫κ
    setView.remove(this);
    
    delete [] _map[0];
    delete [] _oldmap[0];
    delete [] _state[0];
    _map[0] = _oldmap[0] = NULL;
    _state[0] = NULL;
    
    delete [] _map;
    delete [] _oldmap;
    delete [] _state;
    _map = _oldmap = NULL;
    _state = NULL;
}

const XsImage * &XsView::map(const Vector &vec) {
    return _map[vec.y][vec.x];
}

const XsImage * &XsView::oldmap(const Vector &vec) {
    return _oldmap[vec.y][vec.x];
}

int &XsView::state(const Vector &vec) {
    return _state[vec.y][vec.x];
}

void XsView::addSprite(XsSprite *s) {
    setSprite.add(s);
}

void XsView::removeSprite(XsSprite *s) {

    s->state = 0;
    if (s->mustDrawBack()) {
	setOldState(s,1);
    }
    
    setSprite.remove(s);
}

void XsView::setState(const XsSprite *s,int v) {
    
    Vector top		= max(mapTop(),s->mapTop());
    Vector bottom	= min(mapBottom(),s->mapBottom());
    ForAbsE(loop,top,bottom) { state(loop) = v; }
}
void XsView::setOldState(const XsSprite *s,int v) {
    
    Vector top		= max(mapTop(),s->oldmapTop());
    Vector bottom	= min(mapBottom(),s->oldmapBottom());
    ForAbsE(loop,top,bottom) { state(loop) = v; }
}

int XsView::getState(const XsSprite *s) {

    int result = 0;
    
    Vector top		= max(mapTop(),s->mapTop());
    Vector bottom	= min(mapBottom(),s->mapBottom());
    ForAbsE(loop,top,bottom) { result |= state(loop); }

    return result;
}

void XsView::refresh(void) {

    XsSprite *tmpSp = NULL;
    ReadIter<XsSprite *> ri(setSprite);
    DepthSortedSprite setMust;
    ReadIter<XsSprite *> rim(setMust);
    
    // ɽ
    ForAbs(loop,zero,_mapSize) {
	if (map(loop)) {
	    map(loop)->draw(*this,loop*tileSize);
	}
    }
    
    // ץ饤ȤΥȡɽ
    ri.head();
    while (ri.existNext()) {
	tmpSp = ri.elementNext();
	ri.next();
	if (tmpSp->state) {
	    setMust.add(tmpSp);
	}
    }
    rim.head();
    while (rim.existNext()) {
	tmpSp = rim.elementNext();
	rim.next();
	tmpSp->image->draw(*this,Vector(tmpSp->locate));
    }
    
    // ɥؤɽ
    XCopyArea(display,
	      _pixmap,window,gcCopy,
	      zero.x , zero.y ,
	      tileSize.x * _mapSize.x , tileSize.y * _mapSize.y ,
	      zero.x , zero.y);

    // եå
    XFlush(display);

    // ץ饤ȤΥեå
    ri.head();
    while (ri.existNext()) {
	tmpSp = ri.elementNext();
	ri.next();
	tmpSp->flushSprite();
    }
    
    // ޥåפι
    memcpy(_oldmap[0],_map[0],
	   sizeof(const XsImage *) * _mapSize.x * _mapSize.y);

    // ֤ι
    memset(_state[0],0,sizeof(int) * mapArray);    
}

void XsView::refreshAll(void) {

    ReadIter<XsView *> ri(setView);
    ri.head();
    while (ri.existNext()) {
	ri.elementNext()->refresh();
	ri.next();
    }
}

void XsView::flushAllView(void) {

    ReadIter<XsView *> ri(setView);
    ri.head();
    while (ri.existNext()) {
	ri.elementNext()->flushView();
	ri.next();
    }
}

void XsView::flushView(void) {

    //\\\\\
    // ѿ
    int spriteNum = setSprite.num();
    SpriteContena con[spriteNum];	// ƥ
    Set <SpriteContena *> setCheck;	// ץåΥץ饤
    DepthSortedContena setMust;		// 褹륹ץ饤

    //\\\\\\\\\\\\\\
    // 
    {
	// ѹսΥե饰򤿤Ƥ
	for (int loop = 0 ; loop < mapArray ; loop++)
	  _state[0][loop] |= (_map[0][loop] != _oldmap[0][loop]);

	// طʤΥå
	ReadIter <XsSprite *> ri(setSprite);
	int i = 0;
	while (ri.existNext()) {
	    XsSprite *sp = ri.elementNext();
	    ri.next();

	    // طʤɬפʤΤΥե饰򤿤Ƥ
	    if (sp->mustDrawBack()) setOldState(sp,1);

	    // ƥʤ˰ܤؤ
	    SpriteContena &c = con[i++];
	    c.sprite = sp;
	    c.top = sp->mapTop();
	    c.bottom = sp->mapBottom();
	}
    }
    
    //\\\\\\\\\\\\\\\\
    // state  sprite
    {
	// ƥʤ齸˰ܤؤ
	for (int i = 0; i < spriteNum; i ++) {
	    SpriteContena *sc = &con[i];
	    XsSprite *sp = sc->sprite;
	    // ɽ֤ǡ
	    if (sp->state) {
		// View˸֤
		if (0 <= sc->bottom.x &&
		    0 <= sc->bottom.y &&
		    sc->top.x < mapSize().x &&
		    sc->top.y < mapSize().y) {
		    if (sp->mustDraw() || getState(sp))
		      // 񤫤ʤ㤤ʤ
		      setMust.add(sc);
		    else
		      // 괺ɬפϤʤ
		      setCheck.add(sc);
		}
	    }
	}
    }

    //\\\\\\\\\\\
    // 
    {
	int result;
	do {
	    
	    // ɤˤäʤä餳Τޤȴ
	    result = 0;
	    
	    // å٤礫Ϥ
	    ReadIter <SpriteContena *> ric(setCheck);
	    while (ric.existNext() && !result) {
		SpriteContena *sc = ric.elementNext();
		ric.next();
		
		// ʤФʤΤĴ٤
		ReadIter <SpriteContena *> rim(setMust);
		while (rim.existNext() && !result) {
		    SpriteContena *sc2 = rim.elementNext();
		    rim.next();

		    if (sc->bottom.x < sc2->top.x ||
			sc->bottom.y < sc2->top.y ||
			sc2->bottom.x < sc->top.x ||
			sc2->bottom.y < sc->top.y) {

			// ŤʤäƤʤ
			
		    } else {
			
			// ä !!! ŤʤäƤ
			result = 1;
			
			// ʤФʤɲ
			setCheck.remove(sc);
			setMust.add(sc);
		    }
		}
	    }
	} while (result);
    }
    
    //\\\\\\\\\\\\\\\\
    // sprite  state
    {
	// 񤫤ʤ㤤ʤץ饤Ȥ
	ReadIter <SpriteContena *> rim(setMust);
	while (rim.existNext()) {
	    XsSprite *sp = rim.elementNext()->sprite;
	    setState(sp, int(sp));
	    rim.next();
        }
    }

    //\\\\\\\\\\\\\
    // ɽڤӹ
    {

	_numChangedTiles = 0;
	_numChangedAreas = 0;

	// ɥؤɽ
	ForAbsOuter(loop,zero,_mapSize) {
	    
	    XsImage *p = 0;	// ե饰
	    Vector left = loop;	// ü
	    Vector width(1,1);	// 

	    ForAbsInner(loop,zero,_mapSize) {

		XsImage *s = (XsImage *)(state(loop) ? map(loop) : NULL);
		
		if (p == s) {
		    width.x ++;
		} else {
		    if (p) {
			_numChangedTiles ++;
			p->draw(*this,tileSize*left,width);
		    }
		    p = s;
		    left = loop;
		    width.x = 1;
		}
	    }
	    if (p) {
		_numChangedTiles ++;
		p->draw(*this,tileSize*left,width);
	    }
	}
	
	// ץ饤Ȥɽ
	ReadIter <SpriteContena *> rim(setMust);
	while (rim.existNext()) {
	    SpriteContena *sc = rim.elementNext();
	    XsSprite *sp = sc->sprite;
	    rim.next();
	    if (sp->state) sp->image->draw(*this,Vector(sp->locate));
	}

	// ɥؤɽ
	ForAbsOuter(loop,zero,_mapSize) {
	    
	    int p = 0;		// ե饰
	    Vector left = loop;	// ü
	    int width = 0;	// 

	    ForAbsInner(loop,zero,_mapSize) {

		if (isFillUnchangedTiles && !state(loop)) {
		    XFillRectangle(display, window, gcBlack,
				   tileSize.x * loop.x, tileSize.y * loop.y,
				   tileSize.x,tileSize.y);
		}

		int s = state(loop);
		
		if ((p && s) || (!p && !s)) {
		    width ++;
		} else {
		    if (p) {
			_numChangedAreas ++;
			XCopyArea(display,
				  _pixmap,window,gcCopy,
				  tileSize.x * left.x	,tileSize.y * left.y,
				  tileSize.x * width	,tileSize.y,
				  tileSize.x * left.x	,tileSize.y * left.y);
		    }
		    p = s;
		    left = loop;
		    width = 1;
		}
	    }
	    if (p) {
		_numChangedAreas ++;
		XCopyArea(display,
			  _pixmap,window,gcCopy,
			  tileSize.x * left.x	,tileSize.y * left.y,
			  tileSize.x * width	,tileSize.y,
			  tileSize.x * left.x	,tileSize.y * left.y);
	    }
	}
	
	XFlush(display);
	
	// ץ饤ȤΥեå
	ReadIter <XsSprite *> ri(setSprite);
	while (ri.existNext()) {
	    ri.elementNext()->flushSprite();
	    ri.next();
	}

	// ޥåפι
	memcpy(_oldmap[0],_map[0],sizeof(const XsImage *) * mapArray);
	
	// ֤ι
	memset(_state[0],0,sizeof(int) * mapArray);
    }

    // Ϫʬν
    if (_redraw) {
	redraw();
    }
}

void XsView::showState(void) {

    cout << "XsView::showState()" << endl;
    ForAbsOuter(i,zero,_mapSize) {
	cout << setw(2) << i.y << ": ";	
	ForAbsInner(i,zero,_mapSize) {
	    if (state(i)) {
		cout << setw(2) << state(i) % 100 << " ";
	    } else {
		cout << setw(2) << ".." << " ";
	    }
	}
	cout << endl;
    }
}

void XsView::redraw(void) {

    while (XCheckTypedWindowEvent(display,window,Expose,&event)) {

	XCopyArea(display,
		  _pixmap,window,gcCopy,
		  event.xexpose.x,
		  event.xexpose.y,
		  event.xexpose.width,
		  event.xexpose.height,
		  event.xexpose.x,
		  event.xexpose.y);
    }
    XFlush(display);
}

const int XsView::numChangedTiles(void) const {
    return _numChangedTiles;
}

const int XsView::numChangedAreas(void) const {
    return _numChangedAreas;
}

const Vector &XsView::mapSize(void) const {
    return _mapSize;
}

const Vector XsView::mapTop(void) const {
    return zero;
}

const Vector XsView::mapBottom(void) const {
    return _mapSize-Vector(1,1);
}
